/*-
 * Copyright (c) 2001 Kazutaka YOKOTA <yokota@zodiac.mech.utsunomiya-u.ac.jp>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD$
 */

#ifdef TEST
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <termios.h>
#include <sys/queue.h>
#else
#include <stand.h>
#include <string.h>
#include <sys/queue.h>

#include "bootstrap.h"
#endif

#ifdef TEST
extern char **environ;
#endif

#ifndef TEST
#define isprint(c)	((c) >= ' ' && (c) < 127)
#define iscntrl(c)	((c) >= 0 && (c) < 32)
#endif

#define BEEP		putchar('\a');
#define min(x, y)	(((x) < (y)) ? (x) : (y))

#define TRUE		1
#define FALSE		0

#define VUC_ID_CATEGORY	10
#define VUC_ID_VARTITLE	20
#define VUC_ID_VARS	21
#define VUC_ID_HINTS	22
#define VUC_ID_TUNABLES	23
#define VUC_ID_HELP	30
#define VUC_ID_LINEBUF	40
#define VUC_ID_IRQ	50
#define VUC_ID_DRQ	51
#define VUC_ID_PORT	52
#define VUC_ID_MEMORY	53
#define VUC_ID_FLAGS	54

#define BACKSPACE	8
#define TAB		9
#define LF		10
#define CR		13
#define SPACE		32
#define ESC		27
#define BACKTAB		256
#define UP		257
#define DOWN		258
#define SCROLLUP	259
#define SCROLLDOWN	260
#define ENTER		LF
#define NEXT		TAB
#define PREV		BACKTAB
#define ABORT		CTRLCHAR('C')
#define CTRLCHAR(c)	((c) & 0x1f)

struct vuc_obj {
	void	(*setfocus)(struct vuc_obj *, int);
	int	(*exec)(struct vuc_obj *);
	void	(*select)(struct vuc_obj *);
	void	(*redraw)(struct vuc_obj *);
	char	*(*getvalue)(struct vuc_obj *);
	void	(*setvalue)(struct vuc_obj *, char *);
	void	(*clear)(struct vuc_obj *);
	void	(*destroy)(struct vuc_obj *);
	int	id;
	int	type;
#define VUC_TEXTOBJ	0
#define VUC_EDITOBJ	1
#define VUC_LISTOBJ	2
	int	flags;
#define	VUC_DISABLED	(1 << 0)
#define	VUC_NOFOCUS	(1 << 1)
	int	focus;
	int	x, y;
	int	cx, cy;
	void	*p;
	TAILQ_ENTRY(vuc_obj)	link;
};

#define vuc_getid(o)		((o)->id)
#define vuc_getobjflags(o)	((o)->flags)
#define vuc_setobjflags(o, f)	((o)->flags = (f))
#define vuc_setfocusf(o, f)	((o)->setfocus = (f))
#define vuc_setexecf(o, f)	((o)->exec = (f))
#define vuc_setselectf(o, f)	((o)->select = (f))
#define vuc_setdrawf(o, f)	((o)->redraw = (f))
#define vuc_setgetvalf(o, f)	((o)->getvalue = (f))
#define vuc_setsetvalf(o, f)	((o)->setvalue = (f))

#if 0
static int lines;
static int rows;
#endif

static TAILQ_HEAD(vuc_obj_head, vuc_obj) obj_list;
static struct vuc_obj *curr_obj;

static void
vuc_clear(void)
{

	printf("\033[H\033[J");
}

static void
vuc_goto(int x, int y)
{

	printf("\033[%d;%dH", y + 1, x + 1);
}

static void
vuc_attrinvert(void)
{

	/*
	printf("\033[30;44m");
	*/
	printf("\033[7m");
}

static void
vuc_attrbold(void)
{

	printf("\033[1m");
}

static void
vuc_attrreset(void)
{

	printf("\033[m");
}

#define MAXPARAMS	10

static int
vuc_getc(void)
{
	int param[MAXPARAMS];
	int esc;
	int c;
	int n;

	esc = 0;
	for (;;) {
		c = getchar();
		switch (esc) {
		case 0:
			switch (c) {
			case ESC:
				esc = ESC;
				break;
			case BACKSPACE:
			case TAB:
				return c;
			case CTRLCHAR('C'):
				return ABORT;
			case CTRLCHAR('P'):
				return UP;
			case CTRLCHAR('N'):
				return DOWN;
			case LF:
			case CR:
				return ENTER;
			default:
				if (iscntrl(c) || isprint(c))
					return c;
				break;
			}
			break;
		case ESC:
			switch (c) {
			case '[':
				esc = '[';
				n = 0;
				param[0] = 0;
				break;
			default:
				esc = 0;
				break;
			}
			break;
		case '[':
			switch (c) {
			case '0': case '1': case '2': case '3':
			case '4': case '5': case '6': case '7':
			case '8': case '9':
				param[n] *= 10;
				param[n] += c - '0';
				break;
			case ';':
				if (++n >= MAXPARAMS)
					esc = 0;
				break;
			case 'A':
				return UP;
			case 'B':
				return DOWN;
			case 'I':
				return SCROLLUP;
			case 'G':
				return SCROLLDOWN;
			case 'Z':
				return BACKTAB;
			case '~':
				switch (param[0]) {
				case 5:
					return SCROLLUP;
				case 6:
					return SCROLLDOWN;
				default:
					esc = 0;
					break;
				}
			default:
				esc = 0;
				break;
			}
			break;
		default:
			esc = 0;
			break;
		}
	}
	/* NOT REACHED */
}

static int
vuc_gets(char *buf, int len, int x, int y)
{
	int c;
	int l;

	l = strlen(buf);
	vuc_goto(x + l, y);

	for (;;) {
		c = vuc_getc();
		switch (c) {
		case ABORT:
			buf[l] = 0;
			return ABORT;
		case ENTER:
		case TAB:
		case BACKTAB:
			buf[l] = 0;
			return c;
		case BACKSPACE:
			if (l <= 0) {
				BEEP;
				break;
			}
			--l;
			printf("\b_\b");
			break;
		case UP:
		case DOWN:
		case SCROLLUP:
		case SCROLLDOWN:
			break;
		default:
			if (l >= len) {
				BEEP;
				break;
			}
			buf[l++] = c;
			putchar(c);
			break;
		}
	}
	/* NOT REACHED */
}

static void
vuc_init(void)
{

	TAILQ_INIT(&obj_list);

#if 0
	lines = 24;
	v = getenv("LINES");
	if (v != NULL) {
		lines = strtol(v, NULL, 0);
		if (lines <= 0)
			lines = 24;
	}
	rows = 80;
#endif
}

static void
vuc_term(void)
{
	struct vuc_obj *obj;

	while (!TAILQ_EMPTY(&obj_list)) {
		obj = TAILQ_FIRST(&obj_list);
		TAILQ_REMOVE(&obj_list, obj, link);
		(*obj->destroy)(obj);
		free(obj);
	}
	TAILQ_INIT(&obj_list);
}

static int
vuc_addobj(struct vuc_obj *obj)
{

	TAILQ_INSERT_TAIL(&obj_list, obj, link);
	return 0;
}

static void
vuc_deleteobj(struct vuc_obj *obj)
{

	if (obj != NULL) {
		TAILQ_REMOVE(&obj_list, obj, link);
		(*obj->destroy)(obj);
		free(obj);
	}
}

static char
*nop(void)
{

	return NULL;
}

static struct vuc_obj
*vuc_newobj(int type)
{
	struct vuc_obj *obj;

	obj = malloc(sizeof(*obj));
	if (obj == NULL)
		return NULL;
	bzero(obj, sizeof(*obj));
	obj->type = type;
	obj->exec = (int (*)(struct vuc_obj *))nop;
	obj->setfocus = (void (*)(struct vuc_obj *, int))nop;
	obj->redraw = (void (*)(struct vuc_obj *))nop;
	obj->select = (void (*)(struct vuc_obj *))nop;
	obj->getvalue = (char *(*)(struct vuc_obj *))nop;
	obj->setvalue = (void (*)(struct vuc_obj *, char *))nop;
	obj->clear = (void (*)(struct vuc_obj *))nop;
	obj->destroy = (void (*)(struct vuc_obj *))nop;
	return obj;
}

static struct vuc_obj
*vuc_getobj(int id)
{
	struct vuc_obj *obj;

	TAILQ_FOREACH(obj, &obj_list, link) {
		if (obj->id == id)
			return obj;
	}
	return NULL;
}

static void
vuc_enable(int id)
{
	struct vuc_obj *obj;

	obj = vuc_getobj(id);
	if (obj == NULL)
		return;
	if (!(obj->flags & VUC_DISABLED))
		return;
	obj->flags &= ~VUC_DISABLED;
	(*obj->redraw)(obj);
}

static void
vuc_disable(int id)
{
	struct vuc_obj *obj;

	obj = vuc_getobj(id);
	if (obj == NULL)
		return;
	if (obj->flags & VUC_DISABLED)
		return;
	obj->flags |= VUC_DISABLED;
	(*obj->redraw)(obj);
}

static struct vuc_obj
*vuc_setfocus(int id)
{
	struct vuc_obj *obj;

	obj = vuc_getobj(id);
	if (obj == NULL)
		return NULL;
	(*curr_obj->setfocus)(curr_obj, FALSE);
	(*curr_obj->redraw)(curr_obj);
	curr_obj = obj;
	(*curr_obj->setfocus)(curr_obj, TRUE);
	(*curr_obj->redraw)(curr_obj);
	return obj;
}

static void
vuc_select(int id)
{
	struct vuc_obj *obj;

	obj = vuc_getobj(id);
	if (obj == NULL)
		return;
	(*obj->select)(obj);
}

static char
*vuc_getvalue(int id)
{
	struct vuc_obj *obj;

	obj = vuc_getobj(id);
	if (obj == NULL)
		return NULL;
	return (*obj->getvalue)(obj);
}

static void
vuc_setvalue(int id, char *s)
{
	struct vuc_obj *obj;

	obj = vuc_getobj(id);
	if (obj == NULL)
		return;
	(*obj->setvalue)(obj, s);
	(*obj->redraw)(obj);
}

static void
vuc_drawobj(int id)
{
	struct vuc_obj *obj;

	obj = vuc_getobj(id);
	if (obj == NULL)
		return;
	(*obj->redraw)(obj);
}

static void
vuc_clearobj(int id)
{
	struct vuc_obj *obj;

	obj = vuc_getobj(id);
	if (obj == NULL)
		return;
	(*obj->clear)(obj);
	(*obj->redraw)(obj);
}

static void
vuc_setpos(int id, int x, int y, int cx, int cy)
{
	struct vuc_obj *obj;

	obj = vuc_getobj(id);
	if (obj == NULL)
		return;
	if (x > 0)
		obj->x = x;
	if (y > 0)
		obj->y = y;
	if (cx > 0)
		obj->cx = cx;
	if (cy > 0)
		obj->cy = cy;
	(*obj->redraw)(obj);
}

static struct vuc_obj
*vuc_prevobj(struct vuc_obj *obj)
{

	if (obj == NULL)
		return NULL;
	obj = TAILQ_PREV(obj, vuc_obj_head, link);
	if (obj == NULL)
		obj = TAILQ_LAST(&obj_list, vuc_obj_head);
	return obj;
}

static struct vuc_obj
*vuc_nextobj(struct vuc_obj *obj)
{

	if (obj == NULL)
		return NULL;
	obj = TAILQ_NEXT(obj, link);
	if (obj == NULL)
		obj = TAILQ_FIRST(&obj_list);
	return obj;
}

static void
vuc_draw(void)
{
	struct vuc_obj *obj;

	vuc_clear();
	TAILQ_FOREACH(obj, &obj_list, link) {
		if (obj->flags & VUC_DISABLED)
			continue;
		(*obj->redraw)(obj);
	}
}

/* text menu object */

static void
vuc_drawtext(struct vuc_obj *text)
{

	if (text->p == NULL)
		return;
	vuc_goto(text->x, text->y);
	if (text->flags & VUC_DISABLED)
		printf("%*s", -text->cx, "");
	else
		printf("%*.*s", -text->cx, text->cx, text->p);
}

static void
vuc_settext(struct vuc_obj *text, char *s)
{

	text->p = s;
}

static void
vuc_cleartext(struct vuc_obj *text)
{

	text->p = NULL;
}

static int
vuc_maketext(int id, char *s, int x, int y, int cx, int flags)
{
	struct vuc_obj *obj;

	obj = vuc_newobj(VUC_TEXTOBJ);
	obj->id = id;
	obj->p = s;
	obj->x = x;
	obj->y = y;
	if (cx > 0)
		obj->cx = cx;
	else
		obj->cx = strlen(s);
	obj->cy = 1;
	obj->flags = flags | VUC_NOFOCUS;
	obj->redraw = vuc_drawtext;
	obj->setvalue = vuc_settext;
	obj->clear = vuc_cleartext;

	vuc_addobj(obj);
	return 0;
}

/* edit menu object */

struct vuc_edit {
	char	*buf;
	int	len;
	int	protect;
};

static int
vuc_execedit(struct vuc_obj *edit)
{
	struct vuc_edit *e;
	int i;

	e = (struct vuc_edit *)edit->p;
	i = min(edit->cx, e->len);
	vuc_attrinvert();
	i = vuc_gets(e->buf + e->protect, i - e->protect, edit->x + e->protect, edit->y);
	vuc_attrreset();
	return i;
}

static void
vuc_drawedit(struct vuc_obj *edit)
{
	struct vuc_edit *e;
	int l;

	e = (struct vuc_edit *)edit->p;
	l = min(edit->cx, e->len);
	vuc_goto(edit->x, edit->y);
	if (edit->flags & VUC_DISABLED) {
		printf("%*s", -l, "");
	} else if (edit->focus) {
		vuc_attrinvert();
		printf("%.*s", l, e->buf);
		l -= strlen(e->buf);
		while (l-- > 0)
			putchar('_');
		vuc_attrreset();
	} else {
		printf("%*.*s", -l, l, e->buf);
	}
}

static void
vuc_focusedit(struct vuc_obj *edit, int focus)
{

	edit->focus = focus;
}

static char
*vuc_getedit(struct vuc_obj *edit)
{
	struct vuc_edit *e;

	e = (struct vuc_edit *)edit->p;
	return e->buf;
}

static void
vuc_setedit(struct vuc_obj *edit, char *s)
{
	struct vuc_edit *e;
	int l;

	e = (struct vuc_edit *)edit->p;
	l = min(edit->cx, e->len);
	strncpy(e->buf, s, l);
	e->buf[l] = 0;
}

static void
vuc_clearedit(struct vuc_obj *edit)
{
	struct vuc_edit *e;

	e = (struct vuc_edit *)edit->p;
	e->buf[0] = 0;
}

static void
vuc_destroyedit(struct vuc_obj *edit)
{
	struct vuc_edit *e;

	e = (struct vuc_edit *)edit->p;
	free(e->buf);
	free(e);
}

static int
vuc_makeedit(int id, int len, int x, int y, int flags)
{
	struct vuc_obj *obj;
	struct vuc_edit *e;

	e = malloc(sizeof(*e));
	if (e == NULL)
		return -1;
	e->buf = malloc(len + 1);
	if (e->buf == NULL) {
		free(e);
		return -1;
	}
	e->len = len;
	e->buf[0] = 0;
	e->protect = 0;

	obj = vuc_newobj(VUC_EDITOBJ);
	obj->id = id;
	obj->p = e;
	obj->x = x;
	obj->y = y;
	obj->cx = len;
	obj->cy = 1;
	obj->flags = flags;
	obj->exec = vuc_execedit;
	obj->redraw = vuc_drawedit;
	obj->setfocus = vuc_focusedit;
	obj->getvalue = vuc_getedit;
	obj->setvalue = vuc_setedit;
	obj->clear = vuc_clearedit;
	obj->destroy = vuc_destroyedit;

	vuc_addobj(obj);

	return 0;
}

static void
vuc_protectedit(int id, int len)
{
	struct vuc_obj *obj;
	struct vuc_edit *e;

	obj = vuc_getobj(id);
	if (obj == NULL)
		return;
	e = (struct vuc_edit *)obj->p;
	e->protect = len;
}

/* list box menu object */

struct vuc_listitem {
	void	*p;
	TAILQ_ENTRY(vuc_listitem)	link;
};

struct vuc_list {
	int	top;
	int	focus;
	int	lastfocus;
	int	items;
	TAILQ_HEAD(, vuc_listitem)	itemlist;
	char	*(*drawitem)(void *p);
	void	(*selectitem)(void *p, int select);
};

static void
vuc_addlistitem(int id, void *p)
{
	struct vuc_obj *list;
	struct vuc_list *l;
	struct vuc_listitem *item;

	list = vuc_getobj(id);
	if (list == NULL)
		return;
	l = list->p;
	item = malloc(sizeof(*item));
	if (item == NULL)
		return;
	item->p = p;
	TAILQ_INSERT_TAIL(&(l->itemlist), item, link);
	++l->items;
}

static int
vuc_deletelistitem(int id, void *p)
{
	struct vuc_obj *list;
	struct vuc_list *l;
	struct vuc_listitem *item;

	list = vuc_getobj(id);
	if (list == NULL)
		return -1;
	l = list->p;
	TAILQ_FOREACH(item, &(l->itemlist), link) {
		if (item->p == p) {
			TAILQ_REMOVE(&(l->itemlist), item, link);
			free(item);
			--l->items;
			l->top = 0;
			l->focus = l->lastfocus = -1;
			return 0;
		}
	}
	return -1;
}

static void
vuc_setlistitemcb(int id, void (*selectcb)(void *p, int s), char *(*drawcb)(void *p))
{
	struct vuc_obj *list;
	struct vuc_list *l;

	list = vuc_getobj(id);
	if (list == NULL)
		return;
	l = list->p;
	l->selectitem = selectcb;
	l->drawitem = drawcb;
}

static int
vuc_focusedlistitem(struct vuc_obj *list)
{
	struct vuc_list *l;

	l = list->p;
	return l->focus;
}

static void
*vuc_getlistitem(struct vuc_obj *list, int i)
{
	struct vuc_list *l;
	struct vuc_listitem *item;

	l = list->p;
	TAILQ_FOREACH(item, &(l->itemlist), link) {
		if (--i < 0)
			return item->p;
	}
	return NULL;
}

static int
vuc_execlist(struct vuc_obj *list)
{
	struct vuc_list *l;
	int c;

	l = (struct vuc_list *)list->p;

	c = vuc_getc();
	switch (c) {
	case ENTER:
	case SPACE:
		return ENTER;
	case SCROLLUP:
		if ((l->focus >= 0) && (l->top > 0)) {
			l->top -= list->cy - 1;
			if (l->top < 0)
				l->top = 0;
			l->focus = l->top;
			if (l->selectitem)
				(*l->selectitem)(vuc_getlistitem(list, l->focus), TRUE);
			(*list->redraw)(list);
		}
		return c;
	case UP:
		if (l->focus > 0) {
			--l->focus;
			if (l->focus < l->top)
				--l->top;
			if (l->selectitem)
				(*l->selectitem)(vuc_getlistitem(list, l->focus), TRUE);
			(*list->redraw)(list);
		}
		return c;
	case SCROLLDOWN:
		if ((l->focus >= 0) && (l->top + list->cy - 1 < l->items)) {
			l->top += list->cy - 1;
			l->focus = l->top;
			if (l->selectitem)
				(*l->selectitem)(vuc_getlistitem(list, l->focus), TRUE);
			(*list->redraw)(list);
		}
		return c;
	case DOWN:
		if ((l->focus >= 0) && (l->focus < l->items - 1)) {
			++l->focus;
			if (l->focus >= l->top + list->cy)
				++l->top;
			if (l->selectitem)
				(*l->selectitem)(vuc_getlistitem(list, l->focus), TRUE);
			(*list->redraw)(list);
		}
		return c;
	default:
		return c;
	}
	/* NOT REACHED */
}

static void
vuc_drawlist(struct vuc_obj *list)
{
	struct vuc_list *l;
	struct vuc_listitem *item;
	int i;

	l = (struct vuc_list *)list->p;

	i = 0;
	TAILQ_FOREACH(item, &(l->itemlist), link) {
		if (i >= l->top)
			break;
		++i;
	}
	while ((i < l->items) && (i - l->top < list->cy)) {
		vuc_goto(list->x, list->y + i - l->top);
		if (l->focus == i) {
			vuc_attrinvert();
			if (!list->focus)
				vuc_attrbold();
		}
		if (l->drawitem)
			printf("%*.*s", -list->cx, list->cx,
			       (*l->drawitem)(item->p));
		else
			printf("%*s\r", -list->cx, "");
		if (l->focus == i)
			vuc_attrreset();
		item = TAILQ_NEXT(item, link);
		++i;
	}
	while (i - l->top < list->cy) {
		vuc_goto(list->x, list->y + i - l->top);
		printf("%*s\r", -list->cx, "");
		++i;
	}
}

static void
vuc_focuslist(struct vuc_obj *list, int focus)
{
	struct vuc_list *l;

	list->focus = focus;
	l = (struct vuc_list *)list->p;
	if (l->items <= 0)
		return;
	if (focus) {
		l->focus = l->lastfocus;
		if (l->focus < 0)
			l->focus = l->top;
		if (l->selectitem)
			(*l->selectitem)(vuc_getlistitem(list, l->focus), TRUE);
	} else {
		if ((l->focus >= 0) && l->selectitem)
			(*l->selectitem)(vuc_getlistitem(list, l->focus), FALSE);
		l->lastfocus = l->focus;
	}
}

static void
vuc_clearlist(struct vuc_obj *list)
{
	struct vuc_list *l;
	struct vuc_listitem *i1;
	struct vuc_listitem *i2;

	l = (struct vuc_list *)list->p;
	i1 = TAILQ_FIRST(&(l->itemlist));
	while (i1) {
		i2 = TAILQ_NEXT(i1, link);
		free(i1);
		i1 = i2;
	}
	TAILQ_INIT(&(l->itemlist));
	l->items = 0;
	l->top = 0;
	l->focus = -1;
	l->lastfocus = -1;
}

static void
vuc_destroylist(struct vuc_obj *list)
{

	vuc_clearlist(list);
	free(list->p);
}

static int
vuc_makelist(int id, int x, int y, int cx, int cy, int flags)
{
	struct vuc_obj *obj;
	struct vuc_list *l;

	l = malloc(sizeof(*l));
	if (l == NULL)
		return -1;
	l->top = 0;
	l->focus = -1;
	l->lastfocus = -1;
	l->items = 0;
	TAILQ_INIT(&(l->itemlist));
	l->drawitem = NULL;
	l->selectitem = NULL;

	obj = vuc_newobj(VUC_LISTOBJ);
	obj->id = id;
	obj->p = l;
	obj->x = x;
	obj->y = y;
	obj->cx = cx;
	obj->cy = cy;
	obj->flags = flags;
	obj->exec = vuc_execlist;
	obj->redraw = vuc_drawlist;
	obj->setfocus = vuc_focuslist;
	obj->clear = vuc_clearlist;
	obj->destroy = vuc_destroylist;

	vuc_addobj(obj);
	return 0;
}

/* var category list box */

struct category {
	char	*s;
	void	(*f)(void);
};

static char
*draw_category(void *p)
{
	struct category *cat;

	cat = (struct category *)p;
	return cat->s;
}

static void
select_category(void *p, int select)
{
	struct category *cat;

	if (!select)
		return;

	vuc_disable(VUC_ID_IRQ);
	vuc_disable(VUC_ID_DRQ);
	vuc_disable(VUC_ID_PORT);
	vuc_disable(VUC_ID_MEMORY);
	vuc_disable(VUC_ID_FLAGS);
	vuc_disable(VUC_ID_LINEBUF);
	vuc_setvalue(VUC_ID_HELP, "[TAB] change field  Q)uit");

	cat = (struct category *)p;
	(*cat->f)();
}

static int
exec_category(struct vuc_obj *list)
{
	int c;

	c = vuc_execlist(list);
	switch (c) {
	case SPACE:
	case ENTER:
		for (list = vuc_nextobj(list);
		     list->flags & (VUC_NOFOCUS | VUC_DISABLED);
		     list = vuc_nextobj(list))
			;
		vuc_setfocus(vuc_getid(list));
		c = -1;
		break;
	}
	return c;
}

/* device hints list box */

static TAILQ_HEAD(, devhints)	hints_list;

struct devhints {
	int	valid;
#define DEVHINT_VALID_IRQ	(1 << 0)
#define DEVHINT_VALID_DRQ	(1 << 1)
#define DEVHINT_VALID_PORT	(1 << 2)
#define DEVHINT_VALID_MEM	(1 << 3)
#define DEVHINT_VALID_FLAGS	(1 << 4)
	char	*desc;
	char	*bus;
	char	*name;
	int	unit;
	long	irq;
	long	drq;
	long	port;
	long	mem;
	long	flags;
	int	disabled;
	int	mandatery;
	TAILQ_ENTRY(devhints)	link;
};

static void
clear_devhints(struct devhints *hint)
{

	hint->valid = 0;
	hint->desc = NULL;
	hint->bus = NULL;
	hint->name = NULL;
	hint->unit = -1;
	hint->irq = -1;
	hint->drq = -1;
	hint->port = -1;
	hint->mem = 0;
	hint->flags = -1;
	hint->disabled = FALSE;
	hint->mandatery = FALSE;
}

#ifdef TEST
#define HINT_PREFIX	"hint_"
#define HINT_PREFIX_LEN	5
#define HINT_FS		'_'
#else
#define HINT_PREFIX	"hint."
#define HINT_PREFIX_LEN	5
#define HINT_FS		'.'
#endif
#define HINT_RES_IRQ	"irq"
#define HINT_RES_DRQ	"drq"
#define HINT_RES_PORT	"port"
#define HINT_RES_MEM	"maddr"
#define HINT_RES_FLAGS	"flags"
#define HINT_RES_DISABLE "disabled"
#define HINT_RES_MANDATERY "mandatery"

#ifdef TEST
static void
add_devhints(char *name)
#else
static void
add_devhints(char *name, char *val)
#endif
{
	struct devhints *hint;
	struct devhints *h;
	char *unit;
	char *res;
#ifdef TEST
	char *val;
#endif
	long n;
	int i;

	unit = strchr(name, HINT_FS);
	if (unit == NULL)
		return;
	++unit;
	n = strtol(unit, NULL, 0);
	res = strchr(unit, HINT_FS);
	if (res == NULL)
		return;
	++res;
#ifdef TEST
	val = strchr(res, '=');
	if (val == NULL)
		return;
	++val;
#endif

	hint = NULL;
	TAILQ_FOREACH(h, &hints_list, link) {
		i = strncmp(h->name, name, unit - name - 1);
		if (i == 0)
			i = h->unit - n;
		if (i == 0) {
			hint = h;
			break;
		} else if (i < 0) {
			continue;
		} else {
			break;
		}
	}
	if (hint == NULL) {
		hint = malloc(sizeof(*hint));
		if (hint == NULL)
			return;
		clear_devhints(hint);
		hint->name = malloc(unit - name);
		if (hint->name == NULL) {
			free(hint);
			return;
		}
		strncpy(hint->name, name, unit - name - 1);
		hint->name[unit - name - 1] = 0;
		hint->unit = n;
		if (h != NULL)
			TAILQ_INSERT_BEFORE(h, hint, link);
		else
			TAILQ_INSERT_TAIL(&hints_list, hint, link);
	}

	if (strncmp("desc", res, val - res - 1) == 0) {
		if (hint->desc)
			free(hint->desc);
		hint->desc = strdup(val);
	} else if (strncmp("at", res, val - res - 1) == 0) {
		if (hint->bus)
			free(hint->bus);
		hint->bus = strdup(val);
	} else if (strncmp(HINT_RES_IRQ, res, val - res - 1) == 0) {
		n = strtol(val, NULL, 0);
		hint->irq = n;
		hint->valid |= DEVHINT_VALID_IRQ;
	} else if (strncmp(HINT_RES_DRQ, res, val - res - 1) == 0) {
		n = strtol(val, NULL, 0);
		hint->drq = n;
		hint->valid |= DEVHINT_VALID_DRQ;
	} else if (strncmp(HINT_RES_PORT, res, val - res - 1) == 0) {
		n = strtol(val, NULL, 0);
		hint->port = n;
		hint->valid |= DEVHINT_VALID_PORT;
	} else if (strncmp(HINT_RES_MEM, res, val - res - 1) == 0) {
		n = strtol(val, NULL, 0);
		hint->mem = n;
		hint->valid |= DEVHINT_VALID_MEM;
	} else if (strncmp(HINT_RES_FLAGS, res, val - res - 1) == 0) {
		n = strtol(val, NULL, 0);
		hint->flags = n;
		hint->valid |= DEVHINT_VALID_FLAGS;
	} else if (strncmp(HINT_RES_DISABLE, res, val - res - 1) == 0) {
		n = strtol(val, NULL, 0);
		if (n && !hint->mandatery)
			hint->disabled = TRUE;
		else
			hint->disabled = FALSE;
	} else if (strncmp(HINT_RES_MANDATERY, res, val - res - 1) == 0) {
		n = strtol(val, NULL, 0);
		if (n) {
			hint->mandatery = TRUE;
			hint->disabled = FALSE;
		} else
			hint->mandatery = FALSE;
	}
}

static void
update_devhints(struct devhints *hint)
{
	char var[80];
	char val[20];
	int res;

	sprintf(var, "hint%c%s%c%d%c",
		 HINT_FS, hint->name, HINT_FS, hint->unit, HINT_FS);
	res = strlen(var);
	if (res >= sizeof(var))
		return;

	sprintf(var + res, "at");
	if (hint->bus != NULL) {
		setenv(var, hint->bus, 1);
	} else {
		unsetenv(var);
	}
	sprintf(var + res, HINT_RES_IRQ);
	if (hint->valid & DEVHINT_VALID_IRQ) {
		sprintf(val, "%ld", hint->irq);
		setenv(var, val, 1);
	} else {
		unsetenv(var);
	}
	sprintf(var + res, HINT_RES_DRQ);
	if (hint->valid & DEVHINT_VALID_DRQ) {
		sprintf(val, "%ld", hint->drq);
		setenv(var, val, 1);
	} else {
		unsetenv(var);
	}
	sprintf(var + res, HINT_RES_PORT);
	if (hint->valid & DEVHINT_VALID_PORT) {
		sprintf(val, "%#lx", hint->port);
		setenv(var, val, 1);
	} else {
		unsetenv(var);
	}
	sprintf(var + res, HINT_RES_MEM);
	if (hint->valid & DEVHINT_VALID_MEM) {
		sprintf(val, "%#lx", hint->mem);
		setenv(var, val, 1);
	} else {
		unsetenv(var);
	}
	sprintf(var + res, HINT_RES_FLAGS);
	if (hint->valid & DEVHINT_VALID_FLAGS) {
		sprintf(val, "%ld", hint->flags);
		setenv(var, val, 1);
	} else {
		unsetenv(var);
	}
	sprintf(var + res, HINT_RES_DISABLE);
	if (hint->disabled) {
		setenv(var, "1", 1);
	} else {
		unsetenv(var);
	}
	sprintf(var + res, HINT_RES_MANDATERY);
	if (hint->mandatery) {
		setenv(var, "1", 1);
	} else {
		unsetenv(var);
	}
}

static void
free_devhints(void)
{
	struct devhints *h1, *h2;

	h1 = TAILQ_FIRST(&hints_list);
	while (h1 != NULL) {
		h2 = TAILQ_NEXT(h1, link);
		if (h1->desc)
			free(h1->desc);
		if (h1->bus)
			free(h1->bus);
		free(h1->name);
		free(h1);
		h1 = h2;
	}
	TAILQ_INIT(&hints_list);
}

static void
get_devhints(struct devhints *hint)
{
	char *s;

	clear_devhints(hint);

	s = vuc_getvalue(VUC_ID_IRQ);
	if (*s) {
		hint->valid |= DEVHINT_VALID_IRQ;
		hint->irq = strtol(s, NULL, 0);
	}

	s = vuc_getvalue(VUC_ID_DRQ);
	if (*s) {
		hint->valid |= DEVHINT_VALID_DRQ;
		hint->drq = strtol(s, NULL, 0);
	}

	s = vuc_getvalue(VUC_ID_PORT);
	if (*s) {
		hint->valid |= DEVHINT_VALID_PORT;
		hint->port = strtol(s, NULL, 0);
	}

	s = vuc_getvalue(VUC_ID_MEMORY);
	if (*s) {
		hint->valid |= DEVHINT_VALID_MEM;
		hint->mem = strtol(s, NULL, 0);
	}

	s = vuc_getvalue(VUC_ID_FLAGS);
	if (*s) {
		hint->valid |= DEVHINT_VALID_FLAGS;
		hint->flags = strtol(s, NULL, 0);
	}
}

static void
set_devhints(struct devhints *hint)
{
	char buf[20];

	if (hint->valid & DEVHINT_VALID_IRQ)
		sprintf(buf, "%ld", hint->irq);
	else
		buf[0] = 0;
	vuc_setvalue(VUC_ID_IRQ, buf);

	if (hint->valid & DEVHINT_VALID_DRQ)
		sprintf(buf, "%ld", hint->drq);
	else
		buf[0] = 0;
	vuc_setvalue(VUC_ID_DRQ, buf);

	if (hint->valid & DEVHINT_VALID_PORT)
		sprintf(buf, "%#lx", hint->port);
	else
		buf[0] = 0;
	vuc_setvalue(VUC_ID_PORT, buf);

	if (hint->valid & DEVHINT_VALID_MEM)
		sprintf(buf, "%#lx", hint->mem);
	else
		buf[0] = 0;
	vuc_setvalue(VUC_ID_MEMORY, buf);

	if (hint->valid & DEVHINT_VALID_FLAGS)
		sprintf(buf, "%#lx", hint->flags);
	else
		buf[0] = 0;
	vuc_setvalue(VUC_ID_FLAGS, buf);
}

static char
*draw_hints(void *p)
{
	static char buf[128];
	struct devhints *hint;

	hint = p;
	sprintf(buf, "%-8s%8s%-2d %c %s",
		 (hint->bus) ? hint->bus : "", hint->name, hint->unit,
		 (hint->mandatery) ? '*' : ((hint->disabled) ? 'D' : 'E'),
		 (hint->desc) ? hint->desc : "");
	return buf;
}

static void
select_hints(void *p, int select)
{
	struct devhints *hint;

	if (!select)
		return;

	vuc_setvalue(VUC_ID_HELP,
"[TAB] change field  [Enter] edit parameters  E)nable  D)isable  Q)uit");

	hint = (struct devhints *)p;
	set_devhints(hint);
}

static int
exec_hints(struct vuc_obj *list)
{
	struct vuc_obj *obj;
	struct devhints *hint;
	int c;
	int i;

	c = vuc_execlist(list);
	switch (c) {
	case SPACE:
	case ENTER:
		vuc_setvalue(VUC_ID_HELP,
"[TAB] change field  [Enter] save parameters  [^C] discard changes");
		obj = vuc_getobj(VUC_ID_IRQ);
		vuc_setobjflags(obj, vuc_getobjflags(obj) & ~VUC_NOFOCUS);
		obj = vuc_getobj(VUC_ID_DRQ);
		vuc_setobjflags(obj, vuc_getobjflags(obj) & ~VUC_NOFOCUS);
		obj = vuc_getobj(VUC_ID_PORT);
		vuc_setobjflags(obj, vuc_getobjflags(obj) & ~VUC_NOFOCUS);
		obj = vuc_getobj(VUC_ID_MEMORY);
		vuc_setobjflags(obj, vuc_getobjflags(obj) & ~VUC_NOFOCUS);
		obj = vuc_getobj(VUC_ID_FLAGS);
		vuc_setobjflags(obj, vuc_getobjflags(obj) & ~VUC_NOFOCUS);
		vuc_setfocus(VUC_ID_IRQ);
		c = -1;
		break;
	case 'E':
	case 'e':
		i = vuc_focusedlistitem(list);
		hint = (struct devhints *)vuc_getlistitem(list, i);
		if ((hint != NULL) && !hint->mandatery && hint->disabled) {
			hint->disabled = FALSE;
			update_devhints(hint);
			vuc_drawobj(list->id);
		}
		c = -1;
		break;
	case 'D':
	case 'd':
		i = vuc_focusedlistitem(list);
		hint = (struct devhints *)vuc_getlistitem(list, i);
		if ((hint != NULL) && !hint->mandatery && !hint->disabled) {
			hint->disabled = TRUE;
			update_devhints(hint);
			vuc_drawobj(list->id);
		}
		c = -1;
		break;
	}
	return c;
}

static void
list_hints(void)
{
	struct vuc_obj *obj;
	struct devhints h;

	vuc_disable(VUC_ID_VARS);
	vuc_disable(VUC_ID_TUNABLES);
	vuc_enable(VUC_ID_HINTS);

	vuc_setvalue(VUC_ID_VARTITLE,
"-- Hints --------------------------------------------------------------------");

	clear_devhints(&h);
	set_devhints(&h);
	vuc_enable(VUC_ID_IRQ);
	obj = vuc_getobj(VUC_ID_IRQ);
	vuc_setobjflags(obj, vuc_getobjflags(obj) | VUC_NOFOCUS);
	vuc_enable(VUC_ID_DRQ);
	obj = vuc_getobj(VUC_ID_DRQ);
	vuc_setobjflags(obj, vuc_getobjflags(obj) | VUC_NOFOCUS);
	vuc_enable(VUC_ID_PORT);
	obj = vuc_getobj(VUC_ID_PORT);
	vuc_setobjflags(obj, vuc_getobjflags(obj) | VUC_NOFOCUS);
	vuc_enable(VUC_ID_MEMORY);
	obj = vuc_getobj(VUC_ID_MEMORY);
	vuc_setobjflags(obj, vuc_getobjflags(obj) | VUC_NOFOCUS);
	vuc_enable(VUC_ID_FLAGS);
	obj = vuc_getobj(VUC_ID_FLAGS);
	vuc_setobjflags(obj, vuc_getobjflags(obj) | VUC_NOFOCUS);
	vuc_disable(VUC_ID_LINEBUF);
}

/* variables list box */

static TAILQ_HEAD(varslist_head, vars)	vars_list, tunables_list;

struct vars {
	char	*s;
	TAILQ_ENTRY(vars)	link;
};

#ifdef TEST
static struct vars
*add_var(struct varslist_head *h, char *s)
#else
static struct vars
*add_var(struct varslist_head *h, char *name, char *val)
#endif
{
	struct vars *var;

#ifdef TEST
	s = strdup(s);
	if (s == NULL)
		return NULL;
#else
	char *s;

	s = malloc(strlen(name) + strlen(val) + 2);
	if (s == NULL)
		return NULL;
	sprintf(s, "%s=%s", name, val);
#endif
	var = malloc(sizeof(*var));
	if (var == NULL) {
		free(s);
		return NULL;
	}
	var->s = s;
	TAILQ_INSERT_TAIL(h, var, link);
	return var;
}

static void
delete_var(struct varslist_head *h, struct vars *var)
{

	TAILQ_REMOVE(h, var, link);
	free(var->s);
	free(var);
}

static void
free_vars(struct varslist_head *h)
{
	struct vars *v1, *v2;

	v1 = TAILQ_FIRST(h);
	while (v1 != NULL) {
		v2 = TAILQ_NEXT(v1, link);
		free(v1->s);
		free(v1);
		v1 = v2;
	}
	TAILQ_INIT(h);
}

static char
*draw_vars(void *p)
{
	struct vars *var;

	var = (struct vars *)p;
	return var->s;
}

static void
select_var(void *p, int select)
{
	struct vars *var;
	int i;

	if (!select)
		return;

	vuc_setvalue(VUC_ID_HELP,
"[TAB] change field  [Enter] edit variable  A)dd  D)elete  Q)uit");

	var = (struct vars *)p;
	vuc_setvalue(VUC_ID_LINEBUF, var->s);
	i = strchr(var->s, '=') - var->s + 1;
	vuc_protectedit(VUC_ID_LINEBUF, i);
}

static int
exec_vars(struct vuc_obj *list)
{
	struct vuc_list *l;
	struct vars *var;
	int c;
	int i;

	c = vuc_execlist(list);
	switch (c) {
	case SPACE:
	case ENTER:
		vuc_setvalue(VUC_ID_HELP,
			     "[Enter] save change  [^C] discard change");
		i = vuc_focusedlistitem(list);
		var = (struct vars *)vuc_getlistitem(list, i);
		vuc_setvalue(VUC_ID_LINEBUF, var->s);
		i = strchr(var->s, '=') - var->s + 1;
		vuc_protectedit(VUC_ID_LINEBUF, i);
		vuc_setfocus(VUC_ID_LINEBUF);
		c = -1;
		break;
	case 'A':
	case 'a':
		vuc_setvalue(VUC_ID_HELP,
			     "[Enter] save change  [^C] discard change");
		l = (struct vuc_list *)list->p;
		l->focus = -1;
		vuc_clearobj(VUC_ID_LINEBUF);
		vuc_protectedit(VUC_ID_LINEBUF, 0);
		vuc_setfocus(VUC_ID_LINEBUF);
		c = -1;
		break;
	case 'D':
	case 'd':
		i = vuc_focusedlistitem(list);
		var = (struct vars *)vuc_getlistitem(list, i);
		*strchr(var->s, '=') = 0;
		unsetenv(var->s);
		vuc_deletelistitem(list->id, var);
		delete_var((vuc_getid(list) == VUC_ID_VARS) ?
			   &vars_list : &tunables_list, var);
		vuc_setfocus(list->id);
		c = -1;
		break;
	}
	return c;
}

static void
list_vars(void)
{
	struct vuc_obj *obj;

	vuc_disable(VUC_ID_TUNABLES);
	vuc_disable(VUC_ID_HINTS);
	vuc_enable(VUC_ID_VARS);

	vuc_clearobj(VUC_ID_LINEBUF);

	vuc_setvalue(VUC_ID_VARTITLE,
"-- Variables ----------------------------------------------------------------");
	vuc_disable(VUC_ID_IRQ);
	vuc_disable(VUC_ID_DRQ);
	vuc_disable(VUC_ID_PORT);
	vuc_disable(VUC_ID_MEMORY);
	vuc_disable(VUC_ID_FLAGS);
	vuc_enable(VUC_ID_LINEBUF);
	obj = vuc_getobj(VUC_ID_LINEBUF);
	vuc_setobjflags(obj, vuc_getobjflags(obj) | VUC_NOFOCUS);
}

/* kernel tunable paramters list box */

static int
tunables(char *s)
{
	static struct {
		char *s;
		int len;
	} prefix[] = {
#ifdef TEST
		{ "kern_", 5 },
		{ "net_", 4 },
		{ "hw_", 3 },
		{ "machdep_", 8 },
#else
		{ "kern.", 5 },
		{ "net.", 4 },
		{ "hw.", 3 },
		{ "machdep.", 8 },
#endif
	};
	int i;

	for (i = 0; i < sizeof(prefix)/sizeof(prefix[0]); ++i) {
		if (strncmp(s, prefix[i].s, prefix[i].len) == 0)
			return 1;
	}
	return 0;
}

static void
list_tunables(void)
{
	struct vuc_obj *obj;

	vuc_disable(VUC_ID_HINTS);
	vuc_disable(VUC_ID_VARS);
	vuc_enable(VUC_ID_TUNABLES);

	vuc_clearobj(VUC_ID_LINEBUF);

	vuc_setvalue(VUC_ID_VARTITLE,
"-- Kernel Tunable Parameters ------------------------------------------------");
	vuc_disable(VUC_ID_IRQ);
	vuc_disable(VUC_ID_DRQ);
	vuc_disable(VUC_ID_PORT);
	vuc_disable(VUC_ID_MEMORY);
	vuc_disable(VUC_ID_FLAGS);
	vuc_enable(VUC_ID_LINEBUF);
	obj = vuc_getobj(VUC_ID_LINEBUF);
	vuc_setobjflags(obj, vuc_getobjflags(obj) | VUC_NOFOCUS);
}

/* hint parameter buffers */

static int
exec_param(struct vuc_obj *edit)
{
	struct vuc_obj *list;
	struct vuc_obj *obj;
	struct devhints *hint;
	struct devhints h;
	int c;
	int i;

	c = vuc_execedit(edit);
	switch (c) {
	case TAB:
		obj = edit;
		for (obj = vuc_nextobj(obj), i = vuc_getid(obj);
		     (i < VUC_ID_IRQ) || (i > VUC_ID_FLAGS) ||
			(obj->flags & (VUC_NOFOCUS | VUC_DISABLED));
		     obj = vuc_nextobj(obj), i = vuc_getid(obj))
		     ;
		vuc_setfocus(vuc_getid(obj));
		c = -1;
		break;
	case BACKTAB:
		obj = edit;
		for (obj = vuc_prevobj(obj), i = vuc_getid(obj);
		     (i < VUC_ID_IRQ) || (i > VUC_ID_FLAGS) ||
			(obj->flags & (VUC_NOFOCUS | VUC_DISABLED));
		     obj = vuc_prevobj(obj), i = vuc_getid(obj))
		     ;
		vuc_setfocus(vuc_getid(obj));
		c = -1;
		break;
	case ENTER:
		list = vuc_getobj(VUC_ID_HINTS);
		i = vuc_focusedlistitem(list);
		if (i >= 0) {
			hint = (struct devhints *)vuc_getlistitem(list, i);
			h = *hint;
			get_devhints(hint);
			hint->desc = h.desc;
			hint->bus = h.bus;
			hint->name = h.name;
			hint->unit = h.unit;
			hint->disabled = h.disabled;
			hint->mandatery = h.mandatery;
			update_devhints(hint);
		}
		/* FALL THROUGH */
	case ABORT:
		obj = vuc_getobj(VUC_ID_IRQ);
		vuc_setobjflags(obj, vuc_getobjflags(obj) | VUC_NOFOCUS);
		obj = vuc_getobj(VUC_ID_DRQ);
		vuc_setobjflags(obj, vuc_getobjflags(obj) | VUC_NOFOCUS);
		obj = vuc_getobj(VUC_ID_PORT);
		vuc_setobjflags(obj, vuc_getobjflags(obj) | VUC_NOFOCUS);
		obj = vuc_getobj(VUC_ID_MEMORY);
		vuc_setobjflags(obj, vuc_getobjflags(obj) | VUC_NOFOCUS);
		obj = vuc_getobj(VUC_ID_FLAGS);
		vuc_setobjflags(obj, vuc_getobjflags(obj) | VUC_NOFOCUS);
		vuc_setfocus(VUC_ID_HINTS);
		c = -1;
		break;
	}
	return c;
}

/* line buffer */

static int
exec_buf(struct vuc_obj *edit)
{
	struct vuc_obj *list;
	struct varslist_head *h;
	struct vars *var;
	char *s;
#ifndef TEST
	char *v;
#endif
	int c;
	int i;

	list = vuc_getobj(VUC_ID_VARS);
	h = &vars_list;
	if (list->flags & VUC_DISABLED) {
		list = vuc_getobj(VUC_ID_TUNABLES);
		h = &tunables_list;
	}

	c = vuc_execedit(edit);
	switch (c) {
	case TAB:
	case BACKTAB:
		c = -1;
		break;
	case SPACE:
	case ENTER:
		c = -1;
		s = vuc_getvalue(edit->id);
		if ((s == NULL) || (*s == 0) || (strchr(s, '=') == NULL)) {
			BEEP;
			break;
		}
		if (putenv(s) != 0) {
			BEEP;
			break;
		}
		i = vuc_focusedlistitem(list);
		if (i >= 0) {
			s = strdup(s);
			if (s != NULL) {
				var = (struct vars *)vuc_getlistitem(list, i);
				free(var->s);
				var->s = s;
			}
		} else {
#ifdef TEST
			var = add_var(h, s);
#else
			v = strchr(s, '=');
			*(v++) = 0;
			var = add_var(h, s, v);
#endif
			if (var != NULL)
				vuc_addlistitem(vuc_getid(list), var);
		}
		/* FALL THROUGH */
	case ABORT:
		vuc_setfocus(vuc_getid(list));
		c = -1;
		break;
	}
	return c;
}

static void
create_objects(void)
{
	static struct category cat[] = {
		{ "Device hints", list_hints },
		{ "Kernel tunable parameters", list_tunables },
		{ "Other variables", list_vars },
	};
	struct vars *var;
	struct devhints *hint;
	int i;

	vuc_maketext(-1, "Configuration Menu", 30, 0, 0, 0);

	vuc_maketext(-1,
"-- Hint Categories ----------------------------------------------------------",
		     0, 1, 0, 0);
	vuc_makelist(VUC_ID_CATEGORY, 2, 2, 75, 5, 0);
	vuc_setexecf(vuc_getobj(VUC_ID_CATEGORY), exec_category);
	vuc_setlistitemcb(VUC_ID_CATEGORY, select_category, draw_category);
	for (i = 0; i < sizeof(cat)/sizeof(cat[0]); ++i)
		vuc_addlistitem(VUC_ID_CATEGORY, &cat[i]);

	vuc_maketext(VUC_ID_VARTITLE,
"-----------------------------------------------------------------------------",
		     0, 7, 0, 0);
	vuc_makelist(VUC_ID_VARS, 2, 8, 75, 10, VUC_DISABLED);
	vuc_setexecf(vuc_getobj(VUC_ID_VARS), exec_vars);
	vuc_setlistitemcb(VUC_ID_VARS, select_var, draw_vars);
	TAILQ_FOREACH(var, &vars_list, link) {
		if (var->s)
			vuc_addlistitem(VUC_ID_VARS, var);
	}

	vuc_makelist(VUC_ID_TUNABLES, 2, 8, 75, 10, VUC_DISABLED);
	vuc_setexecf(vuc_getobj(VUC_ID_TUNABLES), exec_vars);
	vuc_setlistitemcb(VUC_ID_TUNABLES, select_var, draw_vars);
	TAILQ_FOREACH(var, &tunables_list, link) {
		if (var->s)
			vuc_addlistitem(VUC_ID_TUNABLES, var);
	}

	vuc_makelist(VUC_ID_HINTS, 2, 8, 75, 10, VUC_DISABLED);
	vuc_setexecf(vuc_getobj(VUC_ID_HINTS), exec_hints);
	vuc_setlistitemcb(VUC_ID_HINTS, select_hints, draw_hints);
	TAILQ_FOREACH(hint, &hints_list, link) {
		vuc_addlistitem(VUC_ID_HINTS, hint);
	}

	vuc_maketext(-1,
"-----------------------------------------------------------------------------",
		     0, 18, 0, 0);
	vuc_maketext(-1, "IRQ:", 	 0, 19, 0, 0);
	vuc_makeedit(VUC_ID_IRQ, 2, 	 4, 19, VUC_NOFOCUS);
	vuc_setexecf(vuc_getobj(VUC_ID_IRQ), exec_param);
	vuc_maketext(-1, "DRQ:", 	 7, 19, 0, 0);
	vuc_makeedit(VUC_ID_DRQ, 2,	11, 19, VUC_NOFOCUS);
	vuc_setexecf(vuc_getobj(VUC_ID_DRQ), exec_param);
	vuc_maketext(-1, "Port:",	14, 19, 0, 0);
	vuc_makeedit(VUC_ID_PORT, 6,	19, 19, VUC_NOFOCUS);
	vuc_setexecf(vuc_getobj(VUC_ID_PORT), exec_param);
	vuc_maketext(-1, "Memory:", 	26, 19, 0, 0);
	vuc_makeedit(VUC_ID_MEMORY, 10,	33, 19, VUC_NOFOCUS);
	vuc_setexecf(vuc_getobj(VUC_ID_MEMORY), exec_param);
	vuc_maketext(-1, "Flags:",	44, 19, 0, 0);
	vuc_makeedit(VUC_ID_FLAGS, 10,	51, 19, VUC_NOFOCUS);
	vuc_setexecf(vuc_getobj(VUC_ID_FLAGS), exec_param);

	vuc_makeedit(VUC_ID_LINEBUF, 77, 0, 20, VUC_DISABLED);
	vuc_setexecf(vuc_getobj(VUC_ID_LINEBUF), exec_buf);

	vuc_maketext(-1,
"-----------------------------------------------------------------------------",
		     0, 21, 0, 0);
	vuc_maketext(VUC_ID_HELP, "Q)uit", 0, 22, 77, 0);
}

#ifdef TEST
static int
cmp(const void *p1, const void *p2)
{

	return strcmp(*(char **)p1, *(char **)p2);
}
#endif

static int
vuserconfig(void)
{
	struct vuc_obj *obj;
#ifdef TEST
	char **p;
	int size;
#else
	struct env_var *p;
	char *v;
#endif
	int quit;

	/* collect vairables to edit */
	TAILQ_INIT(&hints_list);
	TAILQ_INIT(&vars_list);
	TAILQ_INIT(&tunables_list);

#ifdef TEST
	for (p = environ, size = 0; *p; ++p)
		++size;
	qsort(environ, size, sizeof(*p), cmp);
	for (p = environ; *p; ++p) {
		if (strncmp(*p, HINT_PREFIX, HINT_PREFIX_LEN) == 0)
			add_devhints(*p + HINT_PREFIX_LEN);
		else if (tunables(*p))
			add_var(&tunables_list, *p);
		else
			add_var(&vars_list, *p);
	}
#else
	for (p = environ; p != NULL; p = p->ev_next) {
		v = getenv(p->ev_name);
		if (strncmp(p->ev_name, HINT_PREFIX, HINT_PREFIX_LEN) == 0)
			add_devhints(p->ev_name + HINT_PREFIX_LEN, v);
		else if (tunables(p->ev_name))
			add_var(&tunables_list, p->ev_name, v);
		else
			add_var(&vars_list, p->ev_name, v);
	}
#endif

	/* create menu objects... */
	vuc_init();
	create_objects();

	/* and set the focus to the first object */
	TAILQ_FOREACH(curr_obj, &obj_list, link) {
		if ((curr_obj->flags & (VUC_NOFOCUS | VUC_DISABLED)) == 0)
			break;
	}
	(*curr_obj->setfocus)(curr_obj, TRUE);

	/* enter main loop */
	vuc_draw();
	quit = FALSE;
	while (!quit) {
		switch ((*curr_obj->exec)(curr_obj)) {
		case ENTER:
			(*curr_obj->select)(curr_obj);
			break;
		case NEXT:
			obj = curr_obj;
			for (obj = vuc_nextobj(obj);
			     obj->flags & (VUC_NOFOCUS | VUC_DISABLED);
			     obj = vuc_nextobj(obj))
				;
			vuc_setfocus(obj->id);
			break;
		case PREV:
			obj = curr_obj;
			for (obj = vuc_prevobj(obj);
			     obj->flags & (VUC_NOFOCUS | VUC_DISABLED);
			     obj = vuc_prevobj(obj))
				;
			vuc_setfocus(obj->id);
			break;
		case 'q':
		case 'Q':
			quit = TRUE;
			break;
		default:
			continue;
		}
	}
	vuc_clear();
	vuc_term();

	free_devhints();
	free_vars(&vars_list);
	free_vars(&tunables_list);

	return 0;
}

#ifdef TEST
int
main(int argc, char *argv[])
{
	struct termios tty;
	struct termios oldtty;
	FILE *fp;
	char **p;

	tcgetattr(0, &tty);
	oldtty = tty;
	tty.c_lflag &= ~(ICANON | ECHO | ISIG);
	tcsetattr(0, TCSANOW, &tty);

	vuserconfig();
	tcsetattr(0, TCSANOW, &oldtty);

	fp = fopen("vuserconfig.out", "w");
	for (p = environ; *p; ++p)
		fprintf(fp, "%s\n", *p);
	fclose(fp);

	exit(0);
}
#else /* !TEST */
/*
 * Edit device hints and other variables
 */
COMMAND_SET(userconfig, "userconfig", "edit device hints and variables",
	    command_userconfig);

static int
command_userconfig(int argc, char *argv[])
{

	vuserconfig();
	return (CMD_OK);
}
#endif /* TEST */
